"""
Abul Naga and Yalcin indexes implementation.

Author: Szymon Moliński (@SimonMolinsky)

Version: 1.0

Last Revision: 2024-02-25

Contributors:
  -

Tutorials:
  -
"""
import numpy as np
from numpy.typing import ArrayLike
from pydantic import field_validator

from surveypie.core import info
from surveypie.structure.index_model import BaseIndex


class AbulNagaYalcinIndex(BaseIndex):
    """
    Abul Naga and Yalcin Indexes metadata

    Attributes
    ----------
    index : float
        Allison-Foster index.

    name : str = 'Abul Naga & Yalcin Index'

    alpha : float, >= 1
        Controls the weight given to inequalities below the median.

    beta : float, >= 1
        Controls the weight given to inequalities above the median.
    """

    alpha: float
    beta: float
    name: str = "Abul Naga & Yalcin Index"

    @field_validator("alpha", "beta")
    @classmethod
    def greater_or_equal_one(cls, v: float) -> float:
        if v < 1:
            raise ValueError(
                'Parameters "alpha" and "beta" must be greater' "or equal to 1"
            )
        return v


def any_index(
    categories: ArrayLike, responses: ArrayLike, alpha=1.0, beta=1.0
) -> AbulNagaYalcinIndex:
    """
    Function for computation Abul Naga & Yalcin index.

    Parameters
    ----------
    categories : array
        Ordered list of possible categories.

    responses : array
        Dataset with ordinal-scale values used for index computation.

    alpha : float, default = 1
        Parameter used for index weighting. Must be greater or equal to 1.
        See Notes to learn more about this parameter.

    beta : float, default = 1
        Parameter used for index weighting. Must be greater or equal to 1.
        See Notes to learn more about this parameter.

    Returns
    -------
    index : AbulNagaYalcinIndex

    Notes
    -----
    With ``alpha`` == ``beta`` inequality is at a minimum when everyone is in
    the same category, and at a maximum when half of the population lies in
    the lowest category and half in the highest category.
    Different calibrations of the parameters and allow the researcher to
    give different weights to inequalities above and below the median of
    the responsiveness distribution -
    for higher values of ``alpha`` (``beta``), less weight is given to
    inequalities below (above) the median.

    References
    ----------
    [1] Abul Naga RH, Yalcin T. Inequality measurement for ordered response
    health data. J Health Econ. 2008 Dec;27(6):1614-25.
    doi: 10.1016/j.jhealeco.2008.07.015. Epub 2008 Aug 19. PMID: 18838185.

    [2] Andrew M. Jones, Nigel Rice, Silvana Robone, Pedro Rosa Dias.
    Inequality and Polarisation in Health Systems’ Responsiveness:
    A Cross- Country Analysis. HEDG Working Paper 10/27, October 2010.
    URL: https://www.york.ac.uk/media/economics/documents/herc/wp/10_27.pdf
    """

    n_categories = len(categories)
    m = int(np.median(responses))
    c = n_categories + 1 - m

    exp_alpha = 0.5**alpha
    exp_beta = 0.5**beta

    k_a_b = (m - 1) * exp_alpha - (1 - (n_categories - m) * exp_beta)

    ds = info(ds=responses, indicators=categories)

    ds = ds["cumulative"]
    p_alpha = ds[ds.index < m]
    p_beta = ds[ds.index >= m]

    # below median
    p_alpha_alpha = p_alpha**alpha
    p_a = np.sum(p_alpha_alpha)

    # above and equal to median
    p_beta_beta = p_beta**beta
    p_b = np.sum(p_beta_beta)

    index = (p_a - p_b + c) / (k_a_b + c)

    return AbulNagaYalcinIndex(
        index=index, alpha=alpha, beta=beta, n_classes=n_categories
    )



"""
Allison-Foster Index implementation.

Author: Marta Leszczyńska (@Reckony)

Version: 1.1

Last Revision: 2024-02-04

Contributors:
  - Szymon Moliński (@SimonMolinsky)

Tutorials:
  - ``allison-foster-index``
"""
from numpy.typing import ArrayLike
from surveypie.structure.index_model import BaseIndex

import numpy


class AllisonFosterIndex(BaseIndex):
    """
    Allison-Foster Index metadata

    Attributes
    ----------
    index : float
        Allison-Foster index.

    name : str = 'Allison-Foster Index'

    u_h : float
        The mean of the values above (H) group median value.

    u_l : float
        The mean of the values below (L) group median value.
    """

    u_h: float
    u_l: float
    name: str = "Allison-Foster Index"


def af_index(data: ArrayLike) -> AllisonFosterIndex:
    """
    Function for computation Allison-Foster index.

    Parameters
    ----------
    data : array like
        Dataset with ordinal-scale values used for index computation.

    Returns
    -------
    BaseIndex : Allison Foster Index

    Notes
    -----
    Index is calculated as a difference between the mean of the values
    above group median value (H) and the mean of the values below (L) group
    median value.

    Algorithm: AF = MEAN(H) - MEAN(L)
    """
    median_value = numpy.median(data)
    values_above_median = [value for value in data if value > median_value]
    values_below_median = [
        value
        for value in data
        if value not in values_above_median and value != median_value
    ]

    mean_h = float(numpy.mean(values_above_median))
    mean_l = float(numpy.mean(values_below_median))
    allison_foster_index = mean_h - mean_l

    return AllisonFosterIndex(
        index=allison_foster_index, u_h=mean_h, u_l=mean_l
    )



import pandas as pd
from numpy.typing import ArrayLike
from typing import Sequence


def info(ds: Sequence, indicators: ArrayLike) -> pd.DataFrame:
    """
    Get basic summary statistics about ordinal dataset.

    Parameters
    ----------
    ds : Sequence
        The collection of answers or grades.

    indicators : Ordered Sequence
        The ordered collection of unique answers or grades.

    Returns
    -------
    summary_df : pandas DataFrame
        Summary statistics for an ordinal dataset.
    """

    # Calculate frequency using pandas value_counts()
    freq_counts = pd.Series(ds).value_counts()
    freq_counts.name = "frequency"
    freq_counts_df = pd.DataFrame(freq_counts)
    freq_counts_df.index.name = "indicator"

    # Calculate ratio
    total_responses = len(ds)
    counts_to_total = freq_counts_df["frequency"] / total_responses
    freq_counts_df["ratio %"] = counts_to_total * 100

    # Add all indicators
    summary_df = pd.DataFrame(
        index=indicators, columns=["frequency", "ratio %"]
    )
    summary_df.index.name = "indicator"

    # Merge the frequency and percent DataFrames
    # to include 0 counts for missing indicators
    summary_df.update(freq_counts_df, join="left")
    summary_df.fillna(0, inplace=True)

    # Calculate cumulative percent
    summary_df["cumulative"] = summary_df["ratio %"].cumsum() / 100

    # Set the last cumulative value to 100
    summary_df.loc[summary_df["cumulative"] > 1, "cumulative"] = 1

    # Create the final DataFrame
    return summary_df


def summary(data: ArrayLike, percentiles=None) -> pd.DataFrame:
    """
    Calculates summary statistics of observed levels
    :param percentiles: the percentiles to include in the output,
    should fall between 0 and 1
    :param data: list-like or array-like object with numerical data
    :return: pandas DataFrame with statistics
    """
    if percentiles is None:
        percentiles = [0.25, 0.5, 0.75]
    df = pd.DataFrame(data).describe(percentiles=percentiles)
    df.columns = ["summary"]
    return df



import numpy as np
from typing import Sequence
from surveypie.core import info
from numpy.typing import ArrayLike

ALPHA_LOOKUP_TABLE = {
    3: round(np.log(2) / np.log(3), 2),
    4: round((np.log(4) - np.log(3) / np.log(2)), 2),
    5: 0.73,
    6: 0.66,
    7: 0.78,
    8: 0.75,
    9: 0.82,
    10: 0.81,
}


def get_apouey_index(categories: ArrayLike, responses: Sequence) -> float:
    number_of_categories = categories[-1]
    if number_of_categories < 3 or number_of_categories > 10:
        raise ValueError(
            "Category out of range. Please use category within range 3-10."
        )

    alpha = ALPHA_LOOKUP_TABLE[number_of_categories]
    proportions_info = info(responses, categories)

    sum_fc_alpha = 0

    for category in range(1, number_of_categories):
        fc = proportions_info["cumulative"][category]
        sum_fc_alpha += abs(fc - 0.5) ** alpha

    apouey_index = 1 - (2**alpha) / categories[-2] * sum_fc_alpha
    return apouey_index
